﻿using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

namespace Kav
{
    public class Renderer
    {
        private GraphicsDevice GraphicsDevice { get; }
        private int RenderDimensionsX { get; }
        private int RenderDimensionsY { get; }

        private RenderTarget2D ShadowRenderTarget { get; }
        private SimpleDepthEffect SimpleDepthEffect { get; }

        private RenderTarget2D gPosition { get; }
        private RenderTarget2D gNormal { get; }
        private RenderTarget2D gAlbedo { get; }
        private RenderTarget2D gMetallicRoughness { get; }
        private RenderTarget2D deferredRenderTarget { get; }

        private RenderTargetBinding[] GBuffer { get; }

        private DeferredPBREffect DeferredPBREffect { get; }

        private SpriteBatch SpriteBatch { get; }

        public Renderer(GraphicsDevice graphicsDevice, int renderDimensionsX, int renderDimensionsY)
        {
            GraphicsDevice = graphicsDevice;
            RenderDimensionsX = renderDimensionsX;
            RenderDimensionsY = renderDimensionsY;

            ShadowRenderTarget = new RenderTarget2D(
                GraphicsDevice,
                1024,
                1024,
                false,
                SurfaceFormat.Single,
                DepthFormat.Depth24
            );

            gPosition = new RenderTarget2D(
                GraphicsDevice,
                renderDimensionsX,
                renderDimensionsY,
                false,
                SurfaceFormat.Vector4,
                DepthFormat.Depth24
            );

            gNormal = new RenderTarget2D(
                GraphicsDevice,
                renderDimensionsX,
                renderDimensionsY,
                false,
                SurfaceFormat.Vector4,
                DepthFormat.None
            );

            gAlbedo = new RenderTarget2D(
                GraphicsDevice,
                renderDimensionsX,
                renderDimensionsY,
                false,
                SurfaceFormat.Color,
                DepthFormat.None
            );

            gMetallicRoughness = new RenderTarget2D(
                GraphicsDevice,
                renderDimensionsX,
                renderDimensionsY,
                false,
                SurfaceFormat.HalfVector2,
                DepthFormat.None
            );

            GBuffer = new RenderTargetBinding[4] {
                new RenderTargetBinding(gPosition),
                new RenderTargetBinding(gNormal),
                new RenderTargetBinding(gAlbedo),
                new RenderTargetBinding(gMetallicRoughness)
            };
            
            deferredRenderTarget = new RenderTarget2D(
                GraphicsDevice,
                renderDimensionsX,
                renderDimensionsY
            );

            SimpleDepthEffect = new SimpleDepthEffect(GraphicsDevice);
            DeferredPBREffect = new DeferredPBREffect(GraphicsDevice);

            SpriteBatch = new SpriteBatch(GraphicsDevice);

            GraphicsDevice.SetRenderTarget(deferredRenderTarget);
            graphicsDevice.Clear(Color.White);
            GraphicsDevice.SetRenderTarget(null);
        }

        public void DeferredRender(
            Camera camera,
            IEnumerable<(Model, Matrix)> modelTransforms,
            IEnumerable<PointLight> pointLights,
            DirectionalLight directionalLight
        ) {
            ShadowMapRender(camera, modelTransforms, directionalLight);

            GraphicsDevice.SetRenderTargets(GBuffer);
            GraphicsDevice.Clear(ClearOptions.Target | ClearOptions.DepthBuffer, Color.Black, 1f, 0);
            GraphicsDevice.DepthStencilState = DepthStencilState.Default;
            GraphicsDevice.BlendState = BlendState.Opaque;

            foreach (var (model, transform) in modelTransforms)
            {
                foreach (var modelMesh in model.Meshes)
                {
                    foreach (var meshPart in modelMesh.MeshParts)
                    {
                        GraphicsDevice.SetVertexBuffer(meshPart.VertexBuffer);
                        GraphicsDevice.Indices = meshPart.IndexBuffer;

                        if (meshPart.Effect is TransformEffect transformEffect)
                        {
                            transformEffect.World = transform;
                            transformEffect.View = camera.View;
                            transformEffect.Projection = camera.Projection;
                        }

                        foreach (var pass in meshPart.Effect.CurrentTechnique.Passes)
                        {
                            pass.Apply();

                            GraphicsDevice.DrawIndexedPrimitives(
                                PrimitiveType.TriangleList,
                                0,
                                0,
                                meshPart.VertexBuffer.VertexCount,
                                0,
                                meshPart.Triangles.Length
                            );
                        }
                    }
                }
            }

            GraphicsDevice.SetRenderTarget(null);
            GraphicsDevice.Clear(Color.CornflowerBlue);

            DeferredPBREffect.GPosition = gPosition;
            DeferredPBREffect.GAlbedo = gAlbedo;
            DeferredPBREffect.GNormal = gNormal;
            DeferredPBREffect.GMetallicRoughness = gMetallicRoughness;
            DeferredPBREffect.ShadowMap = ShadowRenderTarget;
            DeferredPBREffect.EyePosition = Matrix.Invert(camera.View).Translation;

            int i = 0;
            foreach (var pointLight in pointLights)
            {
                if (i > DeferredPBREffect.MaxPointLights) { break; }
                DeferredPBREffect.PointLights[i] = pointLight;
                i++;
            }

            DeferredPBREffect.DirectionalLightColor = directionalLight.Color.ToVector3() * directionalLight.Intensity;
            DeferredPBREffect.DirectionalLightDirection = directionalLight.Direction;

            SpriteBatch.Begin(SpriteSortMode.Deferred, BlendState.NonPremultiplied, null, null, null, DeferredPBREffect);
            SpriteBatch.Draw(deferredRenderTarget, Vector2.Zero, Color.White);
            SpriteBatch.End();
        }

        // for shadow mapping
        public void ShadowMapRender(Camera camera, IEnumerable<(Model, Matrix)> modelTransforms, DirectionalLight directionalLight)
        {
            GraphicsDevice.SetRenderTarget(ShadowRenderTarget);
            GraphicsDevice.Clear(Color.White);
            GraphicsDevice.DepthStencilState = DepthStencilState.Default;
            GraphicsDevice.BlendState = BlendState.Opaque;
            
            var right = Vector3.Cross(Vector3.Up, directionalLight.Direction);
            var up = Vector3.Cross(directionalLight.Direction, right);
            var cameraBoundingFrustum = new BoundingFrustum(camera.View * camera.Projection);
            Vector3[] frustumCorners = cameraBoundingFrustum.GetCorners();

            Vector3 frustumCenter = Vector3.Zero;
            for (var i = 0; i < frustumCorners.Length; i++)
            {
                frustumCenter += frustumCorners[i];
            }
            frustumCenter /= 8f;

            var lightView = Matrix.CreateLookAt(frustumCenter, frustumCenter - directionalLight.Direction, camera.View.Right);

            for (var i = 0; i < frustumCorners.Length; i++)
            {
                frustumCorners[i] = Vector3.Transform(frustumCorners[i], lightView);
            }

            BoundingBox lightBox = BoundingBox.CreateFromPoints(frustumCorners);
            Vector3 boxSize = lightBox.Max - lightBox.Min;
            Vector3 halfBoxSize = boxSize * 0.5f;

            Vector3 lightPosition = frustumCenter + directionalLight.Direction * -lightBox.Min.Z;
            //lightPosition.Z = lightBox.Min.Z;
            //lightPosition = Vector3.Transform(lightPosition, Matrix.Invert(lightRotation));
            
            SimpleDepthEffect.View = Matrix.CreateLookAt(lightPosition, frustumCenter, camera.View.Right);
            SimpleDepthEffect.Projection = Matrix.CreateOrthographicOffCenter(lightBox.Min.X, lightBox.Max.X, lightBox.Min.Y, lightBox.Max.Y, 0, lightBox.Max.Z - lightBox.Min.Z);

            //SimpleDepthEffect.View = directionalLight.View;
            //SimpleDepthEffect.Projection = directionalLight.Projection;
            DeferredPBREffect.LightSpaceMatrix = SimpleDepthEffect.View * SimpleDepthEffect.Projection;

            var globalShadowView = Matrix.CreateLookAt(frustumCenter + directionalLight.Direction * -0.5f, frustumCenter, camera.View.Right);
            var globalShadowProjection = Matrix.CreateOrthographic(1f, 1f, 0f, 1f);
            //var texScaleBias = Matrix.CreateScale(0.5f, -0.5f, 1f) * Matrix.CreateTranslation(0.5f, 0.5f, 0f);
            //DeferredPBREffect.LightSpaceMatrix = globalShadowView * globalShadowProjection; // * texScaleBias;
            
            foreach (var (model, transform) in modelTransforms)
            {
                foreach (var modelMesh in model.Meshes)
                {
                    foreach (var meshPart in modelMesh.MeshParts)
                    {
                        GraphicsDevice.SetVertexBuffer(meshPart.VertexBuffer);
                        GraphicsDevice.Indices = meshPart.IndexBuffer;

                        SimpleDepthEffect.Model = transform;

                        foreach (var pass in SimpleDepthEffect.CurrentTechnique.Passes)
                        {
                            pass.Apply();

                            GraphicsDevice.DrawIndexedPrimitives(
                                PrimitiveType.TriangleList,
                                0,
                                0,
                                meshPart.VertexBuffer.VertexCount,
                                0,
                                meshPart.Triangles.Length
                            );
                        }
                    }
                }
            }
        }

        public void Render(
            Camera camera, 
            IEnumerable<(Model, Matrix)> modelTransforms, 
            IEnumerable<PointLight> pointLights,
            IEnumerable<DirectionalLight> directionalLights
        ) {
            Render(camera.View, camera.Projection, modelTransforms, pointLights, directionalLights);
        }

        private void Render(
            Matrix view,
            Matrix projection,
            IEnumerable<(Model, Matrix)> modelTransforms,
            IEnumerable<PointLight> pointLights,
            IEnumerable<DirectionalLight> directionalLights
        ) {
            foreach (var (model, transform) in modelTransforms)
            {
                foreach (var modelMesh in model.Meshes)
                {
                    foreach (var meshPart in modelMesh.MeshParts)
                    {
                        GraphicsDevice.SetVertexBuffer(meshPart.VertexBuffer);
                        GraphicsDevice.Indices = meshPart.IndexBuffer;

                        if (meshPart.Effect is TransformEffect transformEffect)
                        {
                            transformEffect.World = transform;
                            transformEffect.View = view;
                            transformEffect.Projection = projection;
                        }

                        if (meshPart.Effect is PointLightEffect pointLightEffect)
                        {
                            int i = 0;
                            foreach (var pointLight in pointLights)
                            {
                                if (i > pointLightEffect.MaxPointLights) { break; }
                                pointLightEffect.PointLights[i] = pointLight;
                                i++;
                            }
                        }

                        foreach (var pass in meshPart.Effect.CurrentTechnique.Passes)
                        {
                            pass.Apply();

                            GraphicsDevice.DrawIndexedPrimitives(
                                PrimitiveType.TriangleList,
                                0,
                                0,
                                meshPart.VertexBuffer.VertexCount,
                                0,
                                meshPart.Triangles.Length
                            );
                        }
                    }
                }
            }
        }
    }
}
